#!/usr/bin/env python
# GOSTIM -- P2P F2P secure IM with GOST-cryptography example
# Copyright (C) 2019 Sergey Matveev <stargrave@stargrave.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from functools import partial
from os import makedirs
from os import path
import argparse
import asyncio
import logging

from pyderasn import ASN1Error
from pyderasn import Choice
from pyderasn import Sequence
from pyderasn import tag_ctxc
from pyderasn import UTF8String


MaxTextLen = 1 << 16
MaxMsgLen = MaxTextLen + 128

OUR_NAME = None
THEIR_NAMES = None
OUT_QUEUES = {}
IN_QUEUES = {}
PEER_ALIVES = {}


class MsgHandshake(Sequence):
    schema = ((
        ("peerName", UTF8String(bounds=(1, 256))),
    ))


class MsgText(Sequence):
    schema = ((
        ("text", UTF8String(bounds=(1, MaxTextLen))),
    ))


class Msg(Choice):
    schema = ((
        ("text", MsgText()),
        ("handshake", MsgHandshake(expl=tag_ctxc(0))),
    ))


async def msg_sender(peer_name: str, writer) -> None:
    in_queue = IN_QUEUES[peer_name]
    while True:
        text = await in_queue.get()
        if text is None:
            break
        writer.write(Msg(("text", MsgText((
            ("text", UTF8String(text)),
        )))).encode())
        try:
            await writer.drain()
        except ConnectionResetError:
            del PEER_ALIVES[peer_name]
            return
        logging.info("%s: sent %d characters message", peer_name, len(text))


async def msg_receiver(msg_text: MsgText, peer_name: str) -> None:
    text = str(msg_text["text"])
    logging.info("%s: received %d characters message", peer_name, len(text))
    await OUT_QUEUES[peer_name].put(text)


async def responder(reader, writer):
    _id = writer.get_extra_info("peername")
    logging.info("%s: connected", _id)
    buf = b""
    msg_expected = "handshake"
    peer_name = None
    while True:
        # Read until we get Msg message {{{
        data = await reader.read(MaxMsgLen)
        if data == b"":
            logging.info("%s: closed connection", _id)
            break
        buf += data
        if len(buf) > MaxMsgLen:
            logging.warning("%s: max buffer size exceeded", _id)
            break
        try:
            msg, tail = Msg().decode(buf)
        except ASN1Error:
            continue
        buf = tail
        # }}}
        if msg.choice != msg_expected:
            logging.warning("%s: unexpected %s message", _id, msg.choice)
            break
        if msg_expected == "text":
            try:
                await msg_receiver(msg.value, peer_name)
            except ValueError as err:
                logging.warning("%s: %s", err)
                break
        # Process Handshake message {{{
        elif msg_expected == "handshake":
            logging.info("%s: got %s message", _id, msg_expected)
            msg_handshake = msg.value
            peer_name = str(msg_handshake["peerName"])
            if peer_name not in THEIR_NAMES:
                logging.warning("unknown peer name: %s", peer_name)
                break
            writer.write(Msg(("handshake", MsgHandshake((
                ("peerName", OUR_NAME),
            )))).encode())
            await writer.drain()
            logging.info("%s: session established: %s", _id, peer_name)
            peer_alive = PEER_ALIVES.pop(peer_name, None)
            if peer_alive is not None:
                peer_alive.close()
                await IN_QUEUES[peer_name].put(None)
            PEER_ALIVES[peer_name] = writer
            asyncio.ensure_future(msg_sender(peer_name, writer))
            msg_expected = "text"
        # }}}
    logging.info("%s: disconnecting", _id)
    if msg_expected == "text":
        IN_QUEUES[peer_name].put(None)
    writer.close()


async def initiator(host, port):
    _id = repr((host, port))
    logging.info("%s: dialing", _id)
    reader, writer = await asyncio.open_connection(host, port)
    # Handshake message {{{
    writer.write(Msg(("handshake", MsgHandshake((
        ("peerName", OUR_NAME),
    )))).encode())
    # }}}
    await writer.drain()

    # Wait for Handshake message {{{
    data = await reader.read(256)
    if data == b"":
        logging.warning("%s: no answer, disconnecting", _id)
        writer.close()
        return
    try:
        msg, _ = Msg().decode(data)
    except ASN1Error:
        logging.warning("%s: undecodable answer, disconnecting", _id)
        writer.close()
        return
    logging.info("%s: got %s message", _id, msg.choice)
    if msg.choice != "handshake":
        logging.warning("%s: unexpected message, disconnecting", _id)
        writer.close()
        return
    # }}}
    msg_handshake = msg.value
    peer_name = str(msg_handshake["peerName"])
    if peer_name not in THEIR_NAMES:
        logging.warning("unknown peer name: %s", peer_name)
        writer.close()
        return
    logging.info("%s: session established: %s", _id, peer_name)
    # Run text message sender, initialize transport decoder {{{
    peer_alive = PEER_ALIVES.pop(peer_name, None)
    if peer_alive is not None:
        peer_alive.close()
        await IN_QUEUES[peer_name].put(None)
    PEER_ALIVES[peer_name] = writer
    asyncio.ensure_future(msg_sender(peer_name, writer))
    # }}}
    buf = b""
    # Wait for test messages {{{
    while True:
        data = await reader.read(MaxMsgLen)
        if data == b"":
            break
        buf += data
        if len(buf) > MaxMsgLen:
            logging.warning("%s: max buffer size exceeded", _id)
            break
        try:
            msg, tail = Msg().decode(buf)
        except ASN1Error:
            continue
        buf = tail
        if msg.choice != "text":
            logging.warning("%s: unexpected %s message", _id, msg.choice)
            break
        try:
            await msg_receiver(msg.value, peer_name)
        except ValueError as err:
            logging.warning("%s: %s", err)
            break
    # }}}
    logging.info("%s: disconnecting: %s", _id, peer_name)
    IN_QUEUES[peer_name].put(None)
    writer.close()


async def unixsock_out_processor(reader, writer, out_queue: asyncio.Queue) -> None:
    while True:
        text = await out_queue.get()
        writer.write(("[%s] %s" % (datetime.now(), text)).encode("utf-8"))
        await writer.drain()


async def unixsock_in_processor(reader, writer, in_queue: asyncio.Queue) -> None:
    while True:
        text = await reader.read(MaxTextLen)
        if text == b"":
            break
        await in_queue.put(text.decode("utf-8"))


async def unixsock_state_processor(reader, writer, peer_name: str) -> None:
    peer_writer = PEER_ALIVES.get(peer_name)
    writer.write(
        b"" if peer_writer is None else (" ".join([
            str(i) for i in peer_writer.get_extra_info("peername")[:2]
        ]).encode("utf-8") + b"\n")
    )
    await writer.drain()
    writer.close()


async def unixsock_conn_processor(reader, writer) -> None:
    data = await reader.read(256)
    writer.close()
    host, port = data.decode("utf-8").split(" ")
    await initiator(host=host, port=int(port))


parser = argparse.ArgumentParser(description="GOSTIM")
parser.add_argument(
    "--our-name",
    required=True,
    help="Our peer name",
)
parser.add_argument(
    "--their-names",
    required=True,
    help="Their peer names, comma-separated",
)
parser.add_argument(
    "--bind",
    default="::1",
    help="Address to listen on",
)
parser.add_argument(
    "--port",
    type=int,
    default=6666,
    help="Port to listen on",
)
args = parser.parse_args()
OUR_NAME = UTF8String(args.our_name)
THEIR_NAMES = set(args.their_names.split(","))

# Start listening on all UNIX sockets {{{
for peer_name in THEIR_NAMES:
    makedirs(peer_name, mode=0o700, exist_ok=True)
    out_queue = asyncio.Queue()
    OUT_QUEUES[peer_name] = out_queue
    asyncio.ensure_future(asyncio.start_unix_server(
        partial(unixsock_out_processor, out_queue=out_queue),
        path.join(peer_name, "out"),
    ))
    in_queue = asyncio.Queue()
    IN_QUEUES[peer_name] = in_queue
    asyncio.ensure_future(asyncio.start_unix_server(
        partial(unixsock_in_processor, in_queue=in_queue),
        path.join(peer_name, "in"),
    ))
    asyncio.ensure_future(asyncio.start_unix_server(
        partial(unixsock_state_processor, peer_name=peer_name),
        path.join(peer_name, "state"),
    ))
asyncio.ensure_future(asyncio.start_unix_server(unixsock_conn_processor, "conn"))
# }}}

logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)s %(asctime)s: %(funcName)s: %(message)s",
)
loop = asyncio.get_event_loop()
server = loop.run_until_complete(asyncio.start_server(responder, args.bind, args.port))
logging.info("Listening on: %s", server.sockets[0].getsockname())
loop.run_forever()
