#!/usr/bin/env python
# GOSTIM -- P2P F2P secure IM with GOST-cryptography example
# Copyright (C) 2019 Sergey Matveev <stargrave@stargrave.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
GOSTIM is peer-to-peer friend-to-friend secure instant messenger with
SIGMA-I handshake protocol. It uses GOST-cryptography algorithms
exclusively.

Identification and authentication are based on long-term 34.10-2012-256
public keys, kept in JSON file::

    {
        "our": {
            "prv": "212...f98",
            "pub": "938...2a1"
        },
        "their": {
            "alice": "d36...fce",
            "bob": "aad...24a"
        }
    }

All interaction (except for log messages written on stderr) is done via
several UNIX domain sockets::

    |-- alice
    |   |-- in
    |   |-- out
    |   `-- state
    |-- bob
    |   |-- in
    |   |-- out
    |   `-- state
    `- conn

:in:
 Write outgoing messages to it.
:out:
 Read incoming messages from it.
:state:
 Outputs empty string if no connection established.
 Host and port otherwise.
:conn:
 Write ``HOST PORT`` string to it, to make outgoing connection.

It is recommended to use
`multitail <https://www.vanheusden.com/multitail/>`__,
`tmux <https://github.com/tmux/tmux/wiki>`__,
`rlwrap <https://github.com/hanslub42/rlwrap>`__,
`socat <http://www.dest-unreach.org/socat/>`__
for convenient usage.

Handshake protocol resembles
`SIGMA-I <http://webee.technion.ac.il/~hugo/sigma-pdf.pdf>`__::

    │                                PubA, CookieA                                 │
    │─────────────────────────────────────────────────────────────────────────────>│
    │                                                                              │
    │                                                                              │
    │PubB, CookieB, Enc((IdB, sign(SignPrvB, (CookieA, CookieB, PubB)), MAC(IdB))) │
    │<─────────────────────────────────────────────────────────────────────────────│
    │                                                                              │
    │                                                                              │
    │                                                                              │
    │       Enc((IdA, sign(SignPrvA, (CookieB, CookieA, PubA)), MAC(IdA)))         │
    │─────────────────────────────────────────────────────────────────────────────>│
    │                                                                              │
    │                                                                              │

Transport messages have simple encryption::

    NONCE += 1
    CIPHERTEXT = CTR(KeyEnc, IV=NONCE, PLAINTEXT)
    MAC = MAC(KeyMac, NONCE || CIPHERTEXT)

Following algorithms are used here:

* GOST R 34.10-2012 256-bit for signatures
* GOST R 34.10-2012 256-bit VKO for key agreement
* GOST R 34.11-2012 256-bit for public key hashing
* GOST R 34.12-2015 128-bit blockcipher (Kuznechik) as a cipher
* GOST R 34.13-2015 CTR mode of operation for encryption
* GOST R 34.13-2015 MAC mode of operation for MAC
* HKDF with 34.11-2012 256-bit as a KDF
"""

from datetime import datetime
from functools import partial
from hmac import compare_digest
from os import makedirs
from os import path
from os import urandom
from sys import exit
from typing import Union
import argparse
import asyncio
import json
import logging

from hkdf import Hkdf
from pyderasn import ASN1Error
from pyderasn import Choice
from pyderasn import hexdec
from pyderasn import hexenc
from pyderasn import Integer
from pyderasn import OctetString
from pyderasn import Sequence
from pyderasn import tag_ctxc
from pygost import gost3410
from pygost.gost3410_vko import kek_34102012256
from pygost.gost3410_vko import ukm_unmarshal
from pygost.gost34112012256 import GOST34112012256
from pygost.gost3412 import GOST3412Kuznechik
from pygost.gost3413 import ctr
from pygost.gost3413 import mac
from pygost.utils import long2bytes


KUZNECHIK_BLOCKSIZE = 16
MaxTextLen = 1 << 16
MaxMsgLen = MaxTextLen + 128

CURVE = gost3410.GOST3410Curve(
    *gost3410.CURVE_PARAMS["GostR3410_2001_CryptoPro_A_ParamSet"]
)
KEY_OUR_SIGN_PRV = None
KEY_OUR_SIGN_PUB = None
KEY_OUR_SIGN_PUB_HASH = None
KEYS = {}
OUT_QUEUES = {}
IN_QUEUES = {}
PEER_ALIVES = {}


class Cookie(OctetString):
    bounds = (16, 16)


class PubKey(OctetString):
    bounds = (64, 64)


class MAC(OctetString):
    bounds = (16, 16)


class MsgHandshake0(Sequence):
    schema = ((
        ("cookieInitiator", Cookie()),
        ("pubKeyInitiator", PubKey()),
    ))


class HandshakeTBS(Sequence):
    schema = ((
        ("cookieTheir", Cookie()),
        ("cookieOur", Cookie()),
        ("pubKeyOur", PubKey()),
    ))


class HandshakeTBE(Sequence):
    schema = ((
        ("identity", OctetString(bounds=(32, 32))),
        ("signature", OctetString(bounds=(64, 64))),
        ("identityMac", MAC()),
    ))


class MsgHandshake1(Sequence):
    schema = ((
        ("cookieResponder", Cookie()),
        ("pubKeyResponder", PubKey()),
        ("ukm", OctetString(bounds=(8, 8))),
        ("ciphertext", OctetString()),
        ("ciphertextMac", MAC()),
    ))


class MsgHandshake2(Sequence):
    schema = ((
        ("ciphertext", OctetString()),
        ("ciphertextMac", MAC()),
    ))


class MsgTextPayload(Sequence):
    schema = ((
        ("nonce", Integer(bounds=(0, float("+inf")))),
        ("ciphertext", OctetString(bounds=(1, MaxTextLen))),
    ))


class MsgText(Sequence):
    schema = ((
        ("payload", MsgTextPayload()),
        ("payloadMac", MAC()),
    ))


class Msg(Choice):
    schema = ((
        ("text", MsgText()),
        ("handshake0", MsgHandshake0(expl=tag_ctxc(0))),
        ("handshake1", MsgHandshake1(expl=tag_ctxc(1))),
        ("handshake2", MsgHandshake2(expl=tag_ctxc(2))),
    ))


def validate_tbe(
        msg_handshake: Union[MsgHandshake1, MsgHandshake2],
        key_mac_identity: bytes,
        key_enc: bytes,
        key_mac: bytes,
        cookie_their: Cookie,
        cookie_our: Cookie,
        pub_key_our: PubKey,
) -> str:
    ciphertext = bytes(msg_handshake["ciphertext"])
    mac_tag = mac(GOST3412Kuznechik(key_mac).encrypt, KUZNECHIK_BLOCKSIZE, ciphertext)
    if not compare_digest(mac_tag, bytes(msg_handshake["ciphertextMac"])):
        raise ValueError("invalid MAC")
    plaintext = ctr(
        GOST3412Kuznechik(key_enc).encrypt,
        KUZNECHIK_BLOCKSIZE,
        ciphertext,
        8 * b"\x00",
    )
    try:
        tbe, _ = HandshakeTBE().decode(plaintext)
    except ASN1Error:
        raise ValueError("can not decode TBE")
    key_sign_pub_hash = bytes(tbe["identity"])
    peer = KEYS.get(key_sign_pub_hash)
    if peer is None:
        raise ValueError("unknown identity")
    mac_tag = mac(
        GOST3412Kuznechik(key_mac_identity).encrypt,
        KUZNECHIK_BLOCKSIZE,
        key_sign_pub_hash,
    )
    if not compare_digest(mac_tag, bytes(tbe["identityMac"])):
        raise ValueError("invalid identity MAC")
    tbs = HandshakeTBS((
        ("cookieTheir", cookie_their),
        ("cookieOur", cookie_our),
        ("pubKeyOur", pub_key_our),
    ))
    if not gost3410.verify(
        CURVE,
        peer["pub"],
        GOST34112012256(tbs.encode()).digest(),
        bytes(tbe["signature"]),
    ):
        raise ValueError("invalid signature")
    return peer["name"]


async def msg_sender(peer_name: str, key_enc: bytes, key_mac: bytes, writer) -> None:
    nonce = 0
    encrypter = GOST3412Kuznechik(key_enc).encrypt
    macer = GOST3412Kuznechik(key_mac).encrypt
    in_queue = IN_QUEUES[peer_name]
    while True:
        text = await in_queue.get()
        if text is None:
            break
        ciphertext = ctr(
            encrypter,
            KUZNECHIK_BLOCKSIZE,
            text.encode("utf-8"),
            long2bytes(nonce, 8),
        )
        payload = MsgTextPayload((
            ("nonce", Integer(nonce)),
            ("ciphertext", OctetString(ciphertext)),
        ))
        mac_tag = mac(macer, KUZNECHIK_BLOCKSIZE, payload.encode())
        writer.write(Msg(("text", MsgText((
            ("payload", payload),
            ("payloadMac", MAC(mac_tag)),
        )))).encode())
        try:
            await writer.drain()
        except ConnectionResetError:
            del PEER_ALIVES[peer_name]
            return
        logging.info("%s: sent %d characters message", peer_name, len(text))
        nonce += 1


async def msg_receiver(
        msg_text: MsgText,
        nonce_expected: int,
        macer,
        encrypter,
        peer_name: str,
) -> None:
    payload = msg_text["payload"]
    if int(payload["nonce"]) != nonce_expected:
        raise ValueError("unexpected nonce value")
    mac_tag = mac(macer, KUZNECHIK_BLOCKSIZE, payload.encode())
    if not compare_digest(mac_tag, bytes(msg_text["payloadMac"])):
        raise ValueError("invalid MAC")
    plaintext = ctr(
        encrypter,
        KUZNECHIK_BLOCKSIZE,
        bytes(payload["ciphertext"]),
        long2bytes(nonce_expected, 8),
    )
    text = plaintext.decode("utf-8")
    logging.info("%s: received %d characters message", peer_name, len(text))
    await OUT_QUEUES[peer_name].put(text)


async def responder(reader, writer):
    _id = writer.get_extra_info("peername")
    logging.info("%s: connected", _id)
    buf = b""
    msg_expected = "handshake0"
    nonce_expected = None
    macer = None
    encrypter = None
    peer_name = None
    while True:
        # Read until we get Msg message {{{
        data = await reader.read(MaxMsgLen)
        if data == b"":
            logging.info("%s: closed connection", _id)
            break
        buf += data
        if len(buf) > MaxMsgLen:
            logging.warning("%s: max buffer size exceeded", _id)
            break
        try:
            msg, tail = Msg().decode(buf)
        except ASN1Error:
            continue
        buf = tail
        # }}}
        if msg.choice != msg_expected:
            logging.warning("%s: unexpected %s message", _id, msg.choice)
            break
        if msg_expected == "text":
            try:
                await msg_receiver(
                    msg.value,
                    nonce_expected,
                    macer,
                    encrypter,
                    peer_name,
                )
            except ValueError as err:
                logging.warning("%s: %s", err)
                break
            nonce_expected += 1
        # Process Handshake 0 message {{{
        elif msg_expected == "handshake0":
            logging.info("%s: got %s message", _id, msg_expected)
            msg_handshake0 = msg.value
            # Generate our ephemeral public key and cookie {{{
            cookie_their = msg_handshake0["cookieInitiator"]
            pub_their_raw = msg_handshake0["pubKeyInitiator"]
            pub_their = gost3410.pub_unmarshal(bytes(pub_their_raw))
            cookie_our = Cookie(urandom(16))
            prv = gost3410.prv_unmarshal(urandom(32))
            pub_our = gost3410.public_key(CURVE, prv)
            ukm_raw = urandom(8)
            ukm = ukm_unmarshal(ukm_raw)
            key_session = kek_34102012256(CURVE, prv, pub_their, ukm, mode=2001)
            # }}}
            # Prepare and sign TBS {{{
            pub_our_raw = PubKey(gost3410.pub_marshal(pub_our))
            tbs = HandshakeTBS((
                ("cookieTheir", cookie_their),
                ("cookieOur", cookie_our),
                ("pubKeyOur", pub_our_raw),
            ))
            signature = gost3410.sign(
                CURVE,
                KEY_OUR_SIGN_PRV,
                GOST34112012256(tbs.encode()).digest(),
            )
            # }}}
            # Authenticate/bind identity, prepare TBE {{{
            kdf = Hkdf(None, key_session, hash=GOST34112012256)
            key_handshake1_mac_identity = kdf.expand(b"handshake1-mac-identity")
            mac_tag = mac(
                GOST3412Kuznechik(key_handshake1_mac_identity).encrypt,
                KUZNECHIK_BLOCKSIZE,
                bytes(KEY_OUR_SIGN_PUB_HASH),
            )
            tbe = HandshakeTBE((
                ("identity", KEY_OUR_SIGN_PUB_HASH),
                ("signature", OctetString(signature)),
                ("identityMac", MAC(mac_tag)),
            ))
            tbe_raw = tbe.encode()
            # }}}
            # Encrypt TBE, prepare and send Handshake 1 message
            key_handshake1_enc = kdf.expand(b"handshake1-enc")
            key_handshake1_mac = kdf.expand(b"handshake1-mac")
            ciphertext = ctr(
                GOST3412Kuznechik(key_handshake1_enc).encrypt,
                KUZNECHIK_BLOCKSIZE,
                tbe_raw,
                8 * b"\x00",
            )
            mac_tag = mac(
                GOST3412Kuznechik(key_handshake1_mac).encrypt,
                KUZNECHIK_BLOCKSIZE,
                ciphertext,
            )
            msg = MsgHandshake1((
                ("cookieResponder", cookie_our),
                ("pubKeyResponder", pub_our_raw),
                ("ukm", OctetString(ukm_raw)),
                ("ciphertext", OctetString(ciphertext)),
                ("ciphertextMac", MAC(mac_tag)),
            ))
            writer.write(Msg(("handshake1", msg)).encode())
            # }}}
            await writer.drain()
            msg_expected = "handshake2"
        # }}}
        # Process Handshake 2 message {{{
        elif msg_expected == "handshake2":
            logging.info("%s: got %s message", _id, msg_expected)
            msg_handshake2 = msg.value
            # Validate Handshake message {{{
            key_handshake2_mac_identity = kdf.expand(b"handshake2-mac-identity")
            key_handshake2_enc = kdf.expand(b"handshake2-enc")
            key_handshake2_mac = kdf.expand(b"handshake2-mac")
            try:
                peer_name = validate_tbe(
                    msg_handshake2,
                    key_handshake2_mac_identity,
                    key_handshake2_enc,
                    key_handshake2_mac,
                    cookie_our,
                    cookie_their,
                    pub_their_raw,
                )
            except ValueError as err:
                logging.warning("%s: %s", _id, err)
                break
            # }}}
            logging.info("%s: session established: %s", _id, peer_name)
            # Run text message sender, initialize transport decoder {{{
            key_initiator_enc = kdf.expand(b"transport-initiator-enc")
            key_initiator_mac = kdf.expand(b"transport-initiator-mac")
            key_responder_enc = kdf.expand(b"transport-responder-enc")
            key_responder_mac = kdf.expand(b"transport-responder-mac")
            peer_alive = PEER_ALIVES.pop(peer_name, None)
            if peer_alive is not None:
                peer_alive.close()
                await IN_QUEUES[peer_name].put(None)
            PEER_ALIVES[peer_name] = writer
            asyncio.ensure_future(msg_sender(
                peer_name,
                key_responder_enc,
                key_responder_mac,
                writer,
            ))
            encrypter = GOST3412Kuznechik(key_initiator_enc).encrypt
            macer = GOST3412Kuznechik(key_initiator_mac).encrypt
            nonce_expected = 0
            msg_expected = "text"
            # }}}
        # }}}
    logging.info("%s: disconnecting", _id)
    if msg_expected == "text":
        IN_QUEUES[peer_name].put(None)
    writer.close()


async def initiator(host, port):
    _id = repr((host, port))
    logging.info("%s: dialing", _id)
    reader, writer = await asyncio.open_connection(host, port)
    # Generate our ephemeral public key and cookie, send Handshake 0 message {{{
    cookie_our = Cookie(urandom(16))
    prv = gost3410.prv_unmarshal(urandom(32))
    pub_our = gost3410.public_key(CURVE, prv)
    pub_our_raw = PubKey(gost3410.pub_marshal(pub_our))
    writer.write(Msg(("handshake0", MsgHandshake0((
        ("cookieInitiator", cookie_our),
        ("pubKeyInitiator", pub_our_raw),
    )))).encode())
    # }}}
    await writer.drain()

    # Wait for Handshake 1 message {{{
    data = await reader.read(256)
    if data == b"":
        logging.warning("%s: no answer, disconnecting", _id)
        writer.close()
        return
    try:
        msg, _ = Msg().decode(data)
    except ASN1Error:
        logging.warning("%s: undecodable answer, disconnecting", _id)
        writer.close()
        return
    logging.info("%s: got %s message", _id, msg.choice)
    if msg.choice != "handshake1":
        logging.warning("%s: unexpected message, disconnecting", _id)
        writer.close()
        return
    # }}}
    msg_handshake1 = msg.value
    # Validate Handshake message {{{
    cookie_their = msg_handshake1["cookieResponder"]
    pub_their_raw = msg_handshake1["pubKeyResponder"]
    pub_their = gost3410.pub_unmarshal(bytes(pub_their_raw))
    ukm_raw = bytes(msg_handshake1["ukm"])
    ukm = ukm_unmarshal(ukm_raw)
    key_session = kek_34102012256(CURVE, prv, pub_their, ukm, mode=2001)
    kdf = Hkdf(None, key_session, hash=GOST34112012256)
    key_handshake1_mac_identity = kdf.expand(b"handshake1-mac-identity")
    key_handshake1_enc = kdf.expand(b"handshake1-enc")
    key_handshake1_mac = kdf.expand(b"handshake1-mac")
    try:
        peer_name = validate_tbe(
            msg_handshake1,
            key_handshake1_mac_identity,
            key_handshake1_enc,
            key_handshake1_mac,
            cookie_our,
            cookie_their,
            pub_their_raw,
        )
    except ValueError as err:
        logging.warning("%s: %s, disconnecting", _id, err)
        writer.close()
        return
    # }}}
    # Prepare and send Handshake 2 message {{{
    tbs = HandshakeTBS((
        ("cookieTheir", cookie_their),
        ("cookieOur", cookie_our),
        ("pubKeyOur", pub_our_raw),
    ))
    signature = gost3410.sign(
        CURVE,
        KEY_OUR_SIGN_PRV,
        GOST34112012256(tbs.encode()).digest(),
    )
    key_handshake2_mac_identity = kdf.expand(b"handshake2-mac-identity")
    mac_tag = mac(
        GOST3412Kuznechik(key_handshake2_mac_identity).encrypt,
        KUZNECHIK_BLOCKSIZE,
        bytes(KEY_OUR_SIGN_PUB_HASH),
    )
    tbe = HandshakeTBE((
        ("identity", KEY_OUR_SIGN_PUB_HASH),
        ("signature", OctetString(signature)),
        ("identityMac", MAC(mac_tag)),
    ))
    tbe_raw = tbe.encode()
    key_handshake2_enc = kdf.expand(b"handshake2-enc")
    key_handshake2_mac = kdf.expand(b"handshake2-mac")
    ciphertext = ctr(
        GOST3412Kuznechik(key_handshake2_enc).encrypt,
        KUZNECHIK_BLOCKSIZE,
        tbe_raw,
        8 * b"\x00",
    )
    mac_tag = mac(
        GOST3412Kuznechik(key_handshake2_mac).encrypt,
        KUZNECHIK_BLOCKSIZE,
        ciphertext,
    )
    writer.write(Msg(("handshake2", MsgHandshake2((
        ("ciphertext", OctetString(ciphertext)),
        ("ciphertextMac", MAC(mac_tag)),
    )))).encode())
    # }}}
    await writer.drain()
    logging.info("%s: session established: %s", _id, peer_name)
    # Run text message sender, initialize transport decoder {{{
    key_initiator_enc = kdf.expand(b"transport-initiator-enc")
    key_initiator_mac = kdf.expand(b"transport-initiator-mac")
    key_responder_enc = kdf.expand(b"transport-responder-enc")
    key_responder_mac = kdf.expand(b"transport-responder-mac")
    peer_alive = PEER_ALIVES.pop(peer_name, None)
    if peer_alive is not None:
        peer_alive.close()
        await IN_QUEUES[peer_name].put(None)
    PEER_ALIVES[peer_name] = writer
    asyncio.ensure_future(msg_sender(
        peer_name,
        key_initiator_enc,
        key_initiator_mac,
        writer,
    ))
    encrypter = GOST3412Kuznechik(key_responder_enc).encrypt
    macer = GOST3412Kuznechik(key_responder_mac).encrypt
    # }}}
    buf = b""
    nonce_expected = 0
    # Wait for test messages {{{
    while True:
        data = await reader.read(MaxMsgLen)
        if data == b"":
            break
        buf += data
        if len(buf) > MaxMsgLen:
            logging.warning("%s: max buffer size exceeded", _id)
            break
        try:
            msg, tail = Msg().decode(buf)
        except ASN1Error:
            continue
        buf = tail
        if msg.choice != "text":
            logging.warning("%s: unexpected %s message", _id, msg.choice)
            break
        try:
            await msg_receiver(
                msg.value,
                nonce_expected,
                macer,
                encrypter,
                peer_name,
            )
        except ValueError as err:
            logging.warning("%s: %s", err)
            break
        nonce_expected += 1
    # }}}
    logging.info("%s: disconnecting: %s", _id, peer_name)
    IN_QUEUES[peer_name].put(None)
    writer.close()


async def unixsock_out_processor(reader, writer, out_queue: asyncio.Queue) -> None:
    while True:
        text = await out_queue.get()
        writer.write(("[%s] %s" % (datetime.now(), text)).encode("utf-8"))
        await writer.drain()


async def unixsock_in_processor(reader, writer, in_queue: asyncio.Queue) -> None:
    while True:
        text = await reader.read(MaxTextLen)
        if text == b"":
            break
        await in_queue.put(text.decode("utf-8"))


async def unixsock_state_processor(reader, writer, peer_name: str) -> None:
    peer_writer = PEER_ALIVES.get(peer_name)
    writer.write(
        b"" if peer_writer is None else (" ".join([
            str(i) for i in peer_writer.get_extra_info("peername")[:2]
        ]).encode("utf-8") + b"\n")
    )
    await writer.drain()
    writer.close()


async def unixsock_conn_processor(reader, writer) -> None:
    data = await reader.read(256)
    writer.close()
    host, port = data.decode("utf-8").split(" ")
    await initiator(host=host, port=int(port))


parser = argparse.ArgumentParser(description="GOSTIM")
parser.add_argument(
    "--keys-gen",
    action="store_true",
    help="Generate JSON with our new keypair",
)
parser.add_argument(
    "--keys",
    default="keys.json",
    required=False,
    help="JSON with our and their keys",
)
parser.add_argument(
    "--bind",
    default="::1",
    help="Address to listen on",
)
parser.add_argument(
    "--port",
    type=int,
    default=6666,
    help="Port to listen on",
)
args = parser.parse_args()

if args.keys_gen:
    prv_raw = urandom(32)
    pub = gost3410.public_key(CURVE, gost3410.prv_unmarshal(prv_raw))
    pub_raw = hexenc(gost3410.pub_marshal(pub))
    print(json.dumps({
        "our": {"prv": hexenc(prv_raw), "pub": pub_raw},
        "their": {"myself": pub_raw},
    }, indent=4))
    exit(0)

# Parse and unmarshal our and their keys {{{
with open(args.keys, "rb") as fd:
    _keys = json.loads(fd.read().decode("utf-8"))
KEY_OUR_SIGN_PRV = gost3410.prv_unmarshal(hexdec(_keys["our"]["prv"]))
_pub = hexdec(_keys["our"]["pub"])
KEY_OUR_SIGN_PUB = gost3410.pub_unmarshal(_pub)
KEY_OUR_SIGN_PUB_HASH = OctetString(GOST34112012256(_pub).digest())
for peer_name, pub_raw in _keys["their"].items():
    _pub = hexdec(pub_raw)
    KEYS[GOST34112012256(_pub).digest()] = {
        "name": peer_name,
        "pub": gost3410.pub_unmarshal(_pub),
    }
# }}}

# Start listening on all UNIX sockets {{{
for peer_name in _keys["their"].keys():
    makedirs(peer_name, mode=0o700, exist_ok=True)
    out_queue = asyncio.Queue()
    OUT_QUEUES[peer_name] = out_queue
    asyncio.ensure_future(asyncio.start_unix_server(
        partial(unixsock_out_processor, out_queue=out_queue),
        path.join(peer_name, "out"),
    ))
    in_queue = asyncio.Queue()
    IN_QUEUES[peer_name] = in_queue
    asyncio.ensure_future(asyncio.start_unix_server(
        partial(unixsock_in_processor, in_queue=in_queue),
        path.join(peer_name, "in"),
    ))
    asyncio.ensure_future(asyncio.start_unix_server(
        partial(unixsock_state_processor, peer_name=peer_name),
        path.join(peer_name, "state"),
    ))
asyncio.ensure_future(asyncio.start_unix_server(unixsock_conn_processor, "conn"))
# }}}

logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)s %(asctime)s: %(funcName)s: %(message)s",
)
loop = asyncio.get_event_loop()
server = loop.run_until_complete(asyncio.start_server(responder, args.bind, args.port))
logging.info("Listening on: %s", server.sockets[0].getsockname())
logging.info("Your public key is: %s", _keys["our"]["pub"])
loop.run_forever()
